@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2006-2015 Red Hat, Inc.
+ * Copyright (c) 2006-2019 Red Hat, Inc.
  *
  * All rights reserved.
  *
@@ -194,6 +194,12 @@ struct join_list_entry {
 	mar_cpg_name_t group_name;
 };
 
+struct join_list_confchg_data {
+	mar_cpg_name_t cpg_group;
+	mar_cpg_address_t join_list[CPG_MEMBERS_MAX];
+	int join_list_entries;
+};
+
 /*
  * Service Interfaces required by service_message_handler struct
  */
@@ -312,7 +318,8 @@ static void do_proc_join(
 	const mar_cpg_name_t *name,
 	uint32_t pid,
 	unsigned int nodeid,
-	int reason);
+	int reason,
+	qb_map_t *group_notify_map);
 
 static void do_proc_leave(
 	const mar_cpg_name_t *name,
@@ -645,9 +652,59 @@ static int notify_lib_totem_membership (
 	return CS_OK;
 }
 
+/*
+ * Helper function for notify_lib_joinlist which prepares member_list using
+ * process_info_list with removed left_list items.
+ * member_list_entries - When not NULL it contains number of member_list entries
+ * member_list - When not NULL it is used as pointer to start of preallocated
+ *               array of members. Pointer is adjusted to the end of array on
+ *               exit.
+ */
+static void notify_lib_joinlist_fill_member_list(
+	const mar_cpg_name_t *group_name,
+	int left_list_entries,
+	const mar_cpg_address_t *left_list,
+	int *member_list_entries,
+	mar_cpg_address_t **member_list)
+{
+	struct qb_list_head *iter;
+	int i;
+
+	if (member_list_entries != NULL) {
+		*member_list_entries = 0;
+	}
+
+	qb_list_for_each(iter, &process_info_list_head) {
+		struct process_info *pi = qb_list_entry (iter, struct process_info, list);
+
+		if (mar_name_compare (&pi->group, group_name) == 0) {
+			int in_left_list = 0;
+
+			for (i = 0; i < left_list_entries; i++) {
+				if (left_list[i].nodeid == pi->nodeid && left_list[i].pid == pi->pid) {
+					in_left_list = 1;
+					break ;
+				}
+			}
+
+			if (!in_left_list) {
+				if (member_list_entries != NULL) {
+					(*member_list_entries)++;
+				}
+
+				if (member_list != NULL) {
+					(*member_list)->nodeid = pi->nodeid;
+					(*member_list)->pid = pi->pid;
+					(*member_list)->reason = CPG_REASON_UNDEFINED;
+					(*member_list)++;
+				}
+			}
+		}
+	}
+}
+
 static int notify_lib_joinlist(
 	const mar_cpg_name_t *group_name,
-	void *conn,
 	int joined_list_entries,
 	mar_cpg_address_t *joined_list,
 	int left_list_entries,
@@ -657,31 +714,19 @@ static int notify_lib_joinlist(
 	int size;
 	char *buf;
 	struct qb_list_head *iter;
-	int count;
+	int member_list_entries;
 	struct res_lib_cpg_confchg_callback *res;
 	mar_cpg_address_t *retgi;
+	int i;
 
-	count = 0;
-
-	qb_list_for_each(iter, &process_info_list_head) {
-		struct process_info *pi = qb_list_entry (iter, struct process_info, list);
-		if (mar_name_compare (&pi->group, group_name) == 0) {
-			int i;
-			int founded = 0;
-
-			for (i = 0; i < left_list_entries; i++) {
-				if (left_list[i].nodeid == pi->nodeid && left_list[i].pid == pi->pid) {
-					founded++;
-				}
-			}
-
-			if (!founded)
-				count++;
-		}
-	}
+	/*
+	 * Find size of member_list (use process_info_list but remove items in left_list)
+	 */
+	notify_lib_joinlist_fill_member_list(group_name, left_list_entries, left_list,
+	    &member_list_entries, NULL);
 
 	size = sizeof(struct res_lib_cpg_confchg_callback) +
-		sizeof(mar_cpg_address_t) * (count + left_list_entries + joined_list_entries);
+		sizeof(mar_cpg_address_t) * (member_list_entries + left_list_entries + joined_list_entries);
 	buf = alloca(size);
 	if (!buf)
 		return CS_ERR_LIBRARY;
@@ -689,68 +734,79 @@ static int notify_lib_joinlist(
 	res = (struct res_lib_cpg_confchg_callback *)buf;
 	res->joined_list_entries = joined_list_entries;
 	res->left_list_entries = left_list_entries;
-	res->member_list_entries = count;
+	res->member_list_entries = member_list_entries;
 	retgi = res->member_list;
 	res->header.size = size;
 	res->header.id = id;
 	res->header.error = CS_OK;
 	memcpy(&res->group_name, group_name, sizeof(mar_cpg_name_t));
 
-	qb_list_for_each(iter, &process_info_list_head) {
-		struct process_info *pi=qb_list_entry (iter, struct process_info, list);
-
-		if (mar_name_compare (&pi->group, group_name) == 0) {
-			int i;
-			int founded = 0;
-
-			for (i = 0;i < left_list_entries; i++) {
-				if (left_list[i].nodeid == pi->nodeid && left_list[i].pid == pi->pid) {
-					founded++;
-				}
-			}
-
-			if (!founded) {
-				retgi->nodeid = pi->nodeid;
-				retgi->pid = pi->pid;
-				retgi++;
-			}
-		}
-	}
+	/*
+	 * Fill res->memberlist. Use process_info_list but remove items in left_list.
+	 */
+	notify_lib_joinlist_fill_member_list(group_name, left_list_entries, left_list,
+	    NULL, &retgi);
 
+	/*
+	 * Fill res->left_list
+	 */
 	if (left_list_entries) {
 		memcpy (retgi, left_list, left_list_entries * sizeof(mar_cpg_address_t));
 		retgi += left_list_entries;
 	}
 
 	if (joined_list_entries) {
+		/*
+		 * Fill res->joined_list
+		 */
 		memcpy (retgi, joined_list, joined_list_entries * sizeof(mar_cpg_address_t));
 		retgi += joined_list_entries;
-	}
 
-	if (conn) {
-		api->ipc_dispatch_send (conn, buf, size);
-	} else {
-	qb_list_for_each(iter, &cpg_pd_list_head) {
-			struct cpg_pd *cpd = qb_list_entry (iter, struct cpg_pd, list);
-			if (mar_name_compare (&cpd->group_name, group_name) == 0) {
-				assert (joined_list_entries <= 1);
-				if (joined_list_entries) {
-					if (joined_list[0].pid == cpd->pid &&
-						joined_list[0].nodeid == api->totem_nodeid_get()) {
+		/*
+		 * Update cpd_state for all local joined processes in group
+		 */
+		for (i = 0; i < joined_list_entries; i++) {
+			if (joined_list[i].nodeid == api->totem_nodeid_get()) {
+				qb_list_for_each(iter, &cpg_pd_list_head) {
+					struct cpg_pd *cpd = qb_list_entry (iter, struct cpg_pd, list);
+					if (joined_list[i].pid == cpd->pid &&
+					    mar_name_compare (&cpd->group_name, group_name) == 0) {
 						cpd->cpd_state = CPD_STATE_JOIN_COMPLETED;
 					}
 				}
-				if (cpd->cpd_state == CPD_STATE_JOIN_COMPLETED ||
-					cpd->cpd_state == CPD_STATE_LEAVE_STARTED) {
+			}
+		}
+	}
 
-					api->ipc_dispatch_send (cpd->conn, buf, size);
-					cpd->transition_counter++;
-				}
-				if (left_list_entries) {
-					if (left_list[0].pid == cpd->pid &&
-						left_list[0].nodeid == api->totem_nodeid_get() &&
-						left_list[0].reason == CONFCHG_CPG_REASON_LEAVE) {
+	/*
+	 * Send notification to all ipc clients joined in group_name
+	 */
+	qb_list_for_each(iter, &cpg_pd_list_head) {
+		struct cpg_pd *cpd = qb_list_entry (iter, struct cpg_pd, list);
+		if (mar_name_compare (&cpd->group_name, group_name) == 0) {
+			if (cpd->cpd_state == CPD_STATE_JOIN_COMPLETED ||
+				cpd->cpd_state == CPD_STATE_LEAVE_STARTED) {
 
+				api->ipc_dispatch_send (cpd->conn, buf, size);
+				cpd->transition_counter++;
+			}
+		}
+	}
+
+	if (left_list_entries) {
+		/*
+		 * Zero internal cpd state for all local processes leaving group
+		 * (this loop is not strictly needed because left_list always either
+		 *  contains exactly one process running on local node or more items
+		 *  but none of them is running on local node)
+		 */
+		for (i = 0; i < joined_list_entries; i++) {
+			if (left_list[i].nodeid == api->totem_nodeid_get() &&
+			    left_list[i].reason == CONFCHG_CPG_REASON_LEAVE) {
+				qb_list_for_each(iter, &cpg_pd_list_head) {
+					struct cpg_pd *cpd = qb_list_entry (iter, struct cpg_pd, list);
+					if (left_list[i].pid == cpd->pid &&
+					    mar_name_compare (&cpd->group_name, group_name) == 0) {
 						cpd->pid = 0;
 						memset (&cpd->group_name, 0, sizeof(cpd->group_name));
 						cpd->cpd_state = CPD_STATE_UNJOINED;
@@ -760,7 +816,6 @@ static int notify_lib_joinlist(
 		}
 	}
 
-
 	/*
 	 * Traverse thru cpds and send totem membership for cpd, where it is not send yet
 	 */
@@ -857,7 +912,7 @@ static void downlist_inform_clients (void)
 		}
 
 		/* send confchg event */
-		notify_lib_joinlist(&group, NULL,
+		notify_lib_joinlist(&group,
 			0, NULL,
 			pcd->left_list_entries,
 			pcd->left_list,
@@ -920,6 +975,11 @@ static void joinlist_inform_clients (void)
 	struct joinlist_msg *stored_msg;
 	struct qb_list_head *iter;
 	unsigned int i;
+	qb_map_t *group_notify_map;
+	qb_map_iter_t *miter;
+	struct join_list_confchg_data *jld;
+
+	group_notify_map = qb_skiplist_create();
 
 	i = 0;
 	qb_list_for_each(iter, &joinlist_messages_head) {
@@ -936,9 +996,20 @@ static void joinlist_inform_clients (void)
 		}
 
 		do_proc_join (&stored_msg->group_name, stored_msg->pid, stored_msg->sender_nodeid,
-			CONFCHG_CPG_REASON_NODEUP);
+			CONFCHG_CPG_REASON_NODEUP, group_notify_map);
 	}
 
+	miter = qb_map_iter_create(group_notify_map);
+	while (qb_map_iter_next(miter, (void **)&jld)) {
+		notify_lib_joinlist(&jld->cpg_group,
+				    jld->join_list_entries, jld->join_list,
+				    0, NULL,
+				    MESSAGE_RES_CPG_CONFCHG_CALLBACK);
+		free(jld);
+	}
+	qb_map_iter_free(miter);
+	qb_map_destroy(group_notify_map);
+
 	joinlist_remove_zombie_pi_entries ();
 }
 
@@ -1015,6 +1086,8 @@ static int cpg_node_joinleave_send (unsigned int pid, const mar_cpg_name_t *grou
 	struct iovec req_exec_cpg_iovec;
 	int result;
 
+	memset(&req_exec_cpg_procjoin, 0, sizeof(req_exec_cpg_procjoin));
+
 	memcpy(&req_exec_cpg_procjoin.group_name, group_name, sizeof(mar_cpg_name_t));
 	req_exec_cpg_procjoin.pid = pid;
 	req_exec_cpg_procjoin.reason = reason;
@@ -1116,13 +1189,15 @@ static void do_proc_join(
 	const mar_cpg_name_t *name,
 	uint32_t pid,
 	unsigned int nodeid,
-	int reason)
+	int reason,
+	qb_map_t *group_notify_map)
 {
 	struct process_info *pi;
 	struct process_info *pi_entry;
 	mar_cpg_address_t notify_info;
 	struct qb_list_head *list;
 	struct qb_list_head *list_to_add = NULL;
+	int size;
 
 	if (process_info_find (name, pid, nodeid) != NULL) {
 		return ;
@@ -1156,10 +1231,24 @@ static void do_proc_join(
 	notify_info.nodeid = nodeid;
 	notify_info.reason = reason;
 
-	notify_lib_joinlist(&pi->group, NULL,
-			    1, &notify_info,
-			    0, NULL,
-			    MESSAGE_RES_CPG_CONFCHG_CALLBACK);
+	if (group_notify_map == NULL) {
+		notify_lib_joinlist(&pi->group,
+				    1, &notify_info,
+				    0, NULL,
+				    MESSAGE_RES_CPG_CONFCHG_CALLBACK);
+	} else {
+		struct join_list_confchg_data *jld = qb_map_get(group_notify_map, pi->group.value);
+		if (jld == NULL) {
+			jld = (struct join_list_confchg_data *)calloc(1, sizeof(struct join_list_confchg_data));
+			memcpy(&jld->cpg_group, &pi->group, sizeof(mar_cpg_name_t));
+			qb_map_put(group_notify_map, jld->cpg_group.value, jld);
+		}
+		size = jld->join_list_entries;
+		jld->join_list[size].nodeid = notify_info.nodeid;
+		jld->join_list[size].pid = notify_info.pid;
+		jld->join_list[size].reason = notify_info.reason;
+		jld->join_list_entries++;
+	}
 }
 
 static void do_proc_leave(
@@ -1176,7 +1265,7 @@ static void do_proc_leave(
 	notify_info.nodeid = nodeid;
 	notify_info.reason = reason;
 
-	notify_lib_joinlist(name, NULL,
+	notify_lib_joinlist(name,
 		0, NULL,
 		1, &notify_info,
 		MESSAGE_RES_CPG_CONFCHG_CALLBACK);
@@ -1196,7 +1285,7 @@ static void message_handler_req_exec_cpg_downlist_old (
 	const void *message,
 	unsigned int nodeid)
 {
-	log_printf (LOGSYS_LEVEL_WARNING, "downlist OLD from node 0x%x",
+	log_printf (LOGSYS_LEVEL_DEBUG, "downlist OLD from node " CS_PRI_NODE_ID,
 		nodeid);
 }
 
@@ -1206,7 +1295,7 @@ static void message_handler_req_exec_cpg_downlist(
 {
 	const struct req_exec_cpg_downlist *req_exec_cpg_downlist = message;
 
-	log_printf (LOGSYS_LEVEL_WARNING, "downlist left_list: %d received",
+	log_printf (LOGSYS_LEVEL_DEBUG, "downlist left_list: %d received",
 			req_exec_cpg_downlist->left_nodes);
 }
 
@@ -1217,14 +1306,14 @@ static void message_handler_req_exec_cpg_procjoin (
 {
 	const struct req_exec_cpg_procjoin *req_exec_cpg_procjoin = message;
 
-	log_printf(LOGSYS_LEVEL_DEBUG, "got procjoin message from cluster node 0x%x (%s) for pid %u",
+	log_printf(LOGSYS_LEVEL_DEBUG, "got procjoin message from cluster node " CS_PRI_NODE_ID " (%s) for pid %u",
 		nodeid,
 		api->totem_ifaces_print(nodeid),
 		(unsigned int)req_exec_cpg_procjoin->pid);
 
 	do_proc_join (&req_exec_cpg_procjoin->group_name,
 		req_exec_cpg_procjoin->pid, nodeid,
-		CONFCHG_CPG_REASON_JOIN);
+		CONFCHG_CPG_REASON_JOIN, NULL);
 }
 
 static void message_handler_req_exec_cpg_procleave (
@@ -1233,7 +1322,7 @@ static void message_handler_req_exec_cpg_procleave (
 {
 	const struct req_exec_cpg_procjoin *req_exec_cpg_procjoin = message;
 
-	log_printf(LOGSYS_LEVEL_DEBUG, "got procleave message from cluster node 0x%x (%s) for pid %u",
+	log_printf(LOGSYS_LEVEL_DEBUG, "got procleave message from cluster node " CS_PRI_NODE_ID " (%s) for pid %u",
 		nodeid,
 		api->totem_ifaces_print(nodeid),
 		(unsigned int)req_exec_cpg_procjoin->pid);
@@ -1254,7 +1343,7 @@ static void message_handler_req_exec_cpg_joinlist (
 	const struct join_list_entry *jle = (const struct join_list_entry *)(message + sizeof(struct qb_ipc_response_header));
 	struct joinlist_msg *stored_msg;
 
-	log_printf(LOGSYS_LEVEL_DEBUG, "got joinlist message from node 0x%x",
+	log_printf(LOGSYS_LEVEL_DEBUG, "got joinlist message from node " CS_PRI_NODE_ID,
 		nodeid);
 
 	while ((const char*)jle < message + res->size) {
@@ -1335,7 +1424,7 @@ static void message_handler_req_exec_cpg_partial_mcast (
 	struct iovec iovec[2];
 	int known_node = 0;
 
-	log_printf(LOGSYS_LEVEL_DEBUG, "Got fragmented message from node %d, size = %d bytes\n", nodeid, msglen);
+	log_printf(LOGSYS_LEVEL_DEBUG, "Got fragmented message from node " CS_PRI_NODE_ID ", size = %d bytes\n", nodeid, msglen);
 
 	res_lib_cpg_mcast.header.id = MESSAGE_RES_CPG_PARTIAL_DELIVER_CALLBACK;
 	res_lib_cpg_mcast.header.size = sizeof(res_lib_cpg_mcast) + msglen;
@@ -1403,7 +1492,8 @@ static int cpg_exec_send_joinlist(void)
 	int count = 0;
 	struct qb_list_head *iter;
 	struct qb_ipc_response_header *res;
- 	char *buf;
+	char *buf;
+	size_t buf_size;
 	struct join_list_entry *jle;
 	struct iovec req_exec_cpg_iovec;
 
@@ -1419,11 +1509,13 @@ static int cpg_exec_send_joinlist(void)
 	if (!count)
 		return 0;
 
-	buf = alloca(sizeof(struct qb_ipc_response_header) + sizeof(struct join_list_entry) * count);
+	buf_size = sizeof(struct qb_ipc_response_header) + sizeof(struct join_list_entry) * count;
+	buf = alloca(buf_size);
 	if (!buf) {
 		log_printf(LOGSYS_LEVEL_WARNING, "Unable to allocate joinlist buffer");
 		return -1;
 	}
+	memset(buf, 0, buf_size);
 
 	jle = (struct join_list_entry *)(buf + sizeof(struct qb_ipc_response_header));
 	res = (struct qb_ipc_response_header *)buf;
@@ -1890,6 +1982,8 @@ static void message_handler_req_lib_cpg_mcast (void *conn, const void *message)
 	}
 
 	if (error == CS_OK) {
+		memset(&req_exec_cpg_mcast, 0, sizeof(req_exec_cpg_mcast));
+
 		req_exec_cpg_mcast.header.size = sizeof(req_exec_cpg_mcast) + msglen;
 		req_exec_cpg_mcast.header.id = SERVICE_ID_MAKE(CPG_SERVICE,
 			MESSAGE_REQ_EXEC_CPG_MCAST);
