@@ -110,13 +110,13 @@ struct totemudp_instance {
 
 	void *context;
 
-	void (*totemudp_deliver_fn) (
+	int (*totemudp_deliver_fn) (
 		void *context,
 		const void *msg,
 		unsigned int msg_len,
 		const struct sockaddr_storage *system_from);
 
-	void (*totemudp_iface_change_fn) (
+	int (*totemudp_iface_change_fn) (
 		void *context,
 		const struct totem_ip_address *iface_address,
 		unsigned int ring_no);
@@ -1136,13 +1136,13 @@ int totemudp_initialize (
 
 	void *context,
 
-	void (*deliver_fn) (
+	int (*deliver_fn) (
 		void *context,
 		const void *msg,
 		unsigned int msg_len,
 		const struct sockaddr_storage *system_from),
 
-	void (*iface_change_fn) (
+	int (*iface_change_fn) (
 		void *context,
 		const struct totem_ip_address *iface_address,
 		unsigned int ring_no),
